/*
  Gaétan Siffert le 22.03.2024
  Donnee GTFS teste sur cette source : https://opentransportdata.swiss/en/cookbook/gtfs/#calendar_datestxt
  Definition du format GTFS :
  https://developers.google.com/transit/gtfs/reference?hl=fr
  https://gtfs.org/schedule/reference
*/
import { parse } from "https://deno.land/std@0.215.0/csv/mod.ts";

import * as types from "./types.ts"

export class GTFS {
    routeToStopTimes: types.RouteToStopTimes;
    stopToRoutes: { [stop_id: string]: Set<string> };
    calendar: types.Calendar[];
    calendarDates: types.CalendarDate[];

    constructor() {
      this.routeToStopTimes = {};
      this.stopToRoutes = {};
      this.calendar = []
      this.calendarDates = []
    }

    async readCsv(filePath: string): Promise<any[]> {
        const csvContent = await Deno.readTextFile(filePath);
        const records = await parse(csvContent, {
            header: true,
            skipFirstRow: true,
        });
        return records;
    }

    async loadStopTimes(filePath: string): Promise<any[]> {
      const records = await this.readCsv(filePath);
      // Transformation des données en types appropriés
      const typedRecords = records.map(record => ({
        trip_id: record.trip_id,
        arrival_time: this.timeToMinutes(record.arrival_time),
        departure_time: this.timeToMinutes(record.departure_time),
        stop_id: record.stop_id,
        stop_sequence: parseInt(record.stop_sequence),
        pickup_type: parseInt(record.pickup_type) as types.PickupType,
        drop_off_type: parseInt(record.drop_off_type) as types.DropOffType,
    }));

    return typedRecords;
  }

    async loadTrips(filePath: string): Promise<any[]> {
      const records = await this.readCsv(filePath);
      // Transformation des données en types appropriés
      const typedRecords = records.map(record => ({
        ...record,
    }));

    return typedRecords;
  }

  async loadCalendar(filePath: string): Promise<any[]> {
    const records = await this.readCsv(filePath);
    // Transformation des données en types appropriés
    const typedRecords = records.map(record => ({
      service_id: record.service_id,
      availabilityDays: [
        parseInt(record.sunday) as types.Availability,
        parseInt(record.monday) as types.Availability,
        parseInt(record.tuesday) as types.Availability,
        parseInt(record.wednesday) as types.Availability,
        parseInt(record.thursday) as types.Availability,
        parseInt(record.friday) as types.Availability,
        parseInt(record.saturday) as types.Availability,
      ],
      start_date: this.convertStringToDate(record.start_date) as Date,
      end_date: this.convertStringToDate(record.end_date) as Date,
  }));

    return typedRecords;
  }

    async loadCalendarDates(filePath: string): Promise<any[]> {
      const records = await this.readCsv(filePath);
      // Transformation des données en types appropriés
      const typedRecords = records.map(record => ({
        service_id: record.service_id,
        date: this.convertStringToDate(record.date) as Date,
        exception_type: parseInt(record.exception_type) as types.ExceptionType,
    }));

    return typedRecords;
  }

    async init(folderPath: string){
      this.calendar = await this.loadCalendar(folderPath + "/calendar.txt") as types.Calendar[]
      this.calendarDates = await this.loadCalendarDates(folderPath + "/calendar_dates.txt") as types.CalendarDate[]
      const tripsRecords = await this.loadTrips(folderPath + "/trips.txt") as types.Trip[]
      const stopTimesRecords = await this.loadStopTimes(folderPath + "/stop_times.txt") as types.StopTime[]


      //s'assurer de trier par ordre croissant de stop_sequence puis de departure_time
      stopTimesRecords.sort((a, b) => {
        const sequenceDiff = a.stop_sequence - b.stop_sequence;
        if (sequenceDiff !== 0) return sequenceDiff;
        return a.departure_time - b.departure_time;
      });

      // stopTimesRecords.forEach(st => console.log(st))

      tripsRecords.forEach(trip => {
        if (!this.routeToStopTimes[trip.route_id]) {
          this.routeToStopTimes[trip.route_id] = {};
        }
    
        if (!this.routeToStopTimes[trip.route_id][trip.service_id]) {
          this.routeToStopTimes[trip.route_id][trip.service_id] = {};
        }
    
        if (!this.routeToStopTimes[trip.route_id][trip.service_id][trip.direction_id]) {
          this.routeToStopTimes[trip.route_id][trip.service_id][trip.direction_id] = {};
        }
      })

      stopTimesRecords.forEach(stopTime => {
        const trip = tripsRecords.find(tripRecord => tripRecord.trip_id === stopTime.trip_id);
        if (trip) {
          // routeToStopTimes
          if (!this.routeToStopTimes[trip.route_id][trip.service_id][trip.direction_id][stopTime.trip_id]) {
            this.routeToStopTimes[trip.route_id][trip.service_id][trip.direction_id][stopTime.trip_id] = [];
          }
          this.routeToStopTimes[trip.route_id][trip.service_id][trip.direction_id][trip.trip_id].push(stopTime);

          // stopToRoutes
          // Vérifiez si le stop_id existe déjà, sinon initialisez avec un Set vide
          if (!this.stopToRoutes[stopTime.stop_id]) {
            this.stopToRoutes[stopTime.stop_id] = new Set();
          }
          // Ajoutez le route_id au Set pour le stop_id donné
          this.stopToRoutes[stopTime.stop_id].add(trip.route_id);
        }
      })

    }

    printRouteToStopTimes() {
      for (const route_id in this.routeToStopTimes) {
        console.log(`Route ID: ${route_id}`);
        const services = this.routeToStopTimes[route_id];
        for (const service_id in services) {
          if (service_id === "TA+xf5e0"){

            console.log(`  Service ID: ${service_id}`);
              const directions = services[service_id];
              for (const direction_id in directions) {
                console.log(`    Direction ID: ${direction_id}`);
                const trips = directions[direction_id];
                for (const trip_id in trips) {
                  console.log(`      Trip ID: ${trip_id}`);
                  const stopTimes = trips[trip_id];
                  for (const stopTime of stopTimes) {  
                    console.log(`        StopTime:`, stopTime);
                  }
                }
              }
          }
        }
      }
    }

    getRoutesForStop(stop_id: string): string[] {
      return this.stopToRoutes[stop_id] ? Array.from(this.stopToRoutes[stop_id]) : [];
    }

    findServiceIdForRouteAndDate(route_id: string, date: Date): string[] {
      const dayOfWeek = date.getDay(); // Dimanche = 0, Lundi = 1, ..., Samedi = 6
      let activeServices = this.calendar.filter(service => 
        service.start_date <= date &&
        service.end_date >= date &&
        service.availabilityDays[dayOfWeek] === types.Availability.Available
      );
    
      // Ajuster pour les exceptions
      this.calendarDates.forEach(calendarDate => {
        if (calendarDate.date.getTime() === date.getTime()) {
          if (calendarDate.exception_type === types.ExceptionType.ServiceAdded) {
            activeServices.push(this.calendar.find(service => service.service_id === calendarDate.service_id)!);
          } else if (calendarDate.exception_type === types.ExceptionType.ServiceRemoved) {
            activeServices = activeServices.filter(service => service.service_id !== calendarDate.service_id);
          }
        }
      });
    
      // Filtrer les services actifs pour la route donnée
      const serviceIdsForRoute = activeServices.filter(service => 
        this.routeToStopTimes[route_id] && Object.values(this.routeToStopTimes[route_id])).map(service => service.service_id);

      if (serviceIdsForRoute.length > 1 || serviceIdsForRoute.length === 0){
        console.warn(`There are ${serviceIdsForRoute.length} service IDS for the route ${route_id} at the date ${date}`)
      }

      return serviceIdsForRoute;
    }

    findNextTrips(route_id: string, service_id: string, stop_id: string, departure_time: string): {[direction_id: string]: types.StopTime[]} {
      const departureTimeInMinutes = this.timeToMinutes(departure_time);
      const stopTimesForRouteAndService = this.routeToStopTimes[route_id]?.[service_id];
      if (!stopTimesForRouteAndService) {
        console.warn(`No data found for route_id: ${route_id} and service_id: ${service_id}`);
        return {};
      }
    
      const result: {[direction_id: string]: types.StopTime[]} = {};
    
      Object.entries(stopTimesForRouteAndService).forEach(([direction_id, trips]) => {
        // Chercher le premier trip contenant le stop_id après le departure_time
        for (const [trip_id, stopTimes] of Object.entries(trips)) {
          const matchingStopTimeIndex = stopTimes.findIndex(st => st.stop_id === stop_id && st.departure_time > departureTimeInMinutes);
          if (matchingStopTimeIndex !== -1) {
            // Au lieu de retourner uniquement le stopTime qui correspond, retourner tout le trip
            result[direction_id] = stopTimes;
            // Une fois qu'un trip correspondant est trouvé pour une direction, passer à la suivante
            break;
          }
        }
      });
    
      return result;
    }
    

    convertStringToDate(dateString: string): Date {
      // Extraire l'année, le mois et le jour d'un string sous format YYYYMMDD
      const year = parseInt(dateString.substring(0, 4), 10);
      const month = parseInt(dateString.substring(4, 6), 10) - 1; // Soustraire 1 car Janvier = 0
      const day = parseInt(dateString.substring(6, 8), 10);
    
      const date = new Date(year, month, day);
    
      return date;
    }

    timeToMinutes(time: string): number {
        const [hours, minutes, _seconds] = time.split(':').map(Number);
        return hours * 60 + minutes;
    }

    minutesToTime(minutes: number): string {
        const h = Math.floor(minutes / 60);
        const m = minutes % 60;
        const hours = h < 10 ? `0${h}` : h.toString();
        const mins = m < 10 ? `0${m}` : m.toString();
        return `${hours}:${mins}:00`; // on assume que les secondes sont toujours 00 pour simplifier
    }

    raptor(sourceStopId: string, targetStopId: string, departureTime: string, departureDate: string) {
      // Convert departure time to minutes since start of service
      const departureTimeInMinutes = this.timeToMinutes(departureTime);

      // Initialize the earliest arrival time at stops
      let T: Map<string, number> = new Map();
      T.set(sourceStopId, departureTimeInMinutes);

      // Variable to store the earliest arrival time at target for each round
      let earliestArrivalTimeAtTarget = Infinity;

      // Marked stops to explore in the next round
      let markedStops: Set<string> = new Set([sourceStopId]);

      // Perform rounds until there are no more stops to explore
      while (markedStops.size > 0) {
          let newMarkedStops: Set<string> = new Set();

          // Explore all marked stops
          markedStops.forEach(stopId => {
              // If this stop has routes serving it
              if (this.stopToRoutes[stopId]) {
                  // Find the next trips for all routes serving this stop
                  for (let routeId of this.stopToRoutes[stopId]) {
                      // Find active service IDs for the date
                      const serviceIds = this.findServiceIdForRouteAndDate(routeId, this.convertStringToDate(departureDate));

                      serviceIds.forEach(serviceId => {
                        const nextTrips = this.findNextTrips(routeId, serviceId, stopId, this.minutesToTime(T.get(stopId) ?? departureTimeInMinutes));
                        Object.entries(nextTrips).forEach(([directionId, stopTimes]) => {
                            // Find the index of the current stop in the stop times
                            const currentIndex = stopTimes.findIndex(st => st.stop_id === stopId);
                    
                            // If the current stop is found, iterate from there onwards
                            if (currentIndex !== -1) {
                                for (let i = currentIndex; i < stopTimes.length; i++) {
                                    let stopTime = stopTimes[i];
                                    // If we haven't reached this stop before or we can reach it earlier
                                    if (!T.has(stopTime.stop_id) || T.get(stopTime.stop_id)! > stopTime.arrival_time) {
                                        T.set(stopTime.stop_id, stopTime.arrival_time);
                                        newMarkedStops.add(stopTime.stop_id);
                    
                                        // If this is the target stop and it's an earlier time, update the earliest time
                                        if (stopTime.stop_id === targetStopId && stopTime.arrival_time < earliestArrivalTimeAtTarget) {
                                            earliestArrivalTimeAtTarget = stopTime.arrival_time;
                                        }
                                    }
                                }
                            }
                        });
                    });
                  }
              }
          });

          // Prepare for the next round
          markedStops = newMarkedStops;
      }

      // If the target was reached, return the earliest arrival time
      if (earliestArrivalTimeAtTarget < Infinity) {
          console.log(`Found trip! Arrival time at target: ${this.minutesToTime(earliestArrivalTimeAtTarget)}`);
          return this.minutesToTime(earliestArrivalTimeAtTarget);
      } else {
          console.log("No trip found within time constraints.");
          return null;
      }
    }
  }
