import copy
from typing import Any
from typing import Dict, List

from altair import Optional, Union
import partridge as ptg
import geopandas as gpd
import pyproj
from shapely.geometry import Point

class StopAccessState:
    def __init__(self, origin_stop_id: str):
        """State tracker for stop ids."""
        self._stops = {}

        # initialize the origin node with no prior trip history
        self._origin = origin_stop_id
        self.try_add_update(self._origin, 0)
    
    def all_stops(self):
        return list(self._stops.keys())
        
    def has_stop(self, stop_id: str):
        return stop_id in self._stops.keys()
    
    def get_stop(self, stop_id: str):
        return self._stops[stop_id]
    
    def try_add_update(
        self,
        stop_id: str,
        time_to_reach: Union[int, float],
        trip_id: Optional[str]=None,
        preceding_path: Optional[List[str]]=None,
    ) -> bool:
        # initialize return object
        did_update = False

        if stop_id in self._stops.keys():
            if self._stops[stop_id]["time_to_reach"] > time_to_reach:
                # update the stop access attributes
                self._stops[stop_id]["time_to_reach"] = time_to_reach
                did_update = True

        else:
            self._stops[stop_id] = {
                "time_to_reach": time_to_reach,
                "preceding": [],  # initialize with no past paths
            }
            did_update = True

        if did_update:
            # override if a preceding path is provided
            if preceding_path:
                self._stops[stop_id]["preceding"] = preceding_path

            # add current trip id to the path of trips taken, avoiding dupes
            if trip_id is not None and (
                len(self._stops[stop_id]["preceding"]) == 0 or
                trip_id != self._stops[stop_id]["preceding"][-1]):
                self._stops[stop_id]["preceding"].append(trip_id)
            
        return did_update


# load a GTFS of AC Transit
path = './gtfs_filtered.zip'
date, service_ids = ptg.read_busiest_date(path)
print(date)
print(service_ids)
view = {'trips.txt': {'service_id': service_ids}}
print(view)
feed = ptg.load_feed(path, view)
print(feed)

# convert all known stops in the schedule to shapes in a GeoDataFrame
gdf = gpd.GeoDataFrame(
    {"stop_id": feed.stops.stop_id.tolist()},
    geometry=[
        Point(lon, lat)
        for lat, lon in zip(
            feed.stops.stop_lat,
            feed.stops.stop_lon)])
gdf = gdf.set_index("stop_id")
gdf.crs = "epsg:4326"

# re-cast to meter-based projection to allow for distance calculations
aeqd = pyproj.Proj(
    proj='aeqd',
    ellps='WGS84',
    datum='WGS84',
    lat_0=gdf.iloc[0].geometry.centroid.y,
    lon_0=gdf.iloc[0].geometry.centroid.x).srs
gdf = gdf.to_crs(crs=aeqd)

# let's use this example origin and destination
# to find the time it would take to go from one to another
from_stop_id = "8588802"  # "Carouge GE, Stade de Genève"
to_stop_id = "8587387"  # "Genève, Bel-Air"

# QA: we know the best way to connect these two is the 51A -> 1T
# if we depart at 8:30 AM, schedule should suggest:
#     take 51A 8:37 - 8:49
#     make walk connection
#     take 1T 8:56 - 9:03
# total travel time: 26 minutes

# look at all trips from that stop that are after the depart time (8h30)
departure_secs = 8.5 * 60 * 60

def get_trip_ids_for_stop(feed, stop_id: str, departure_time: int):
    """Takes a stop and departure time and get associated trip ids."""
    mask_1 = feed.stop_times.stop_id == stop_id
    mask_2 = feed.stop_times.departure_time >= departure_time

    # extract the list of qualifying trip ids
    potential_trips = feed.stop_times[mask_1 & mask_2].trip_id.unique().tolist()
    return potential_trips

def stop_times_for_kth_trip(
    stops_state: StopAccessState,
    last_updated_stops: List[str],
) -> None:
    # sort stops into their associated groups
    trip_stop_pairings = {}
    for ref_stop_id in last_updated_stops:
        # find all trips already related to this stop
        associated_trips = stops_state.get_stop(ref_stop_id)["preceding"]

        # find all qualifying trips assocaited with this stop
        potential_trips = get_trip_ids_for_stop(feed, ref_stop_id, departure_secs)
        for potential_trip in potential_trips:
            # pass on trips that are already addressed
            if potential_trip in associated_trips:
                continue
                
            if potential_trip in trip_stop_pairings.keys():
                trip_stop_pairings[potential_trip].append(ref_stop_id)
            else:
                trip_stop_pairings[potential_trip] = [ref_stop_id]

    # iterate through trips with grouped stops in them
    for trip_id in trip_stop_pairings:
        stop_ids = trip_stop_pairings[trip_id]

        # get all the stop time arrivals for that trip
        stop_times_sub = feed.stop_times[feed.stop_times.trip_id == trip_id]
        stop_times_sub = stop_times_sub.sort_values(by="stop_sequence", ascending=True)
        
        # find all stop ids that are in this stop ordering and pick last on route path
        stop_times_mask = stop_times_sub.stop_id.isin(stop_ids)
        target_stops = stop_times_sub[stop_times_mask].sort_values(by="stop_sequence", ascending=True)
        
        # get the "hop on" point
        from_here = target_stops.tail(1).squeeze()
        ref_stop_id = from_here.stop_id
        
        # are we continuing from some previous path of trips?
        ref_stop_state = stops_state.get_stop(ref_stop_id)
        preceding_path = ref_stop_state["preceding"]

        # how long it took to get to the stop so far (0 for start node)
        baseline_cost = ref_stop_state["time_to_reach"]

        # get all following stops
        stop_times_after_mask = stop_times_sub.stop_sequence >= from_here.stop_sequence
        stop_times_after = stop_times_sub[stop_times_after_mask]

        # for all following stops, calculate time to reach
        arrivals_zip = zip(stop_times_after.arrival_time, stop_times_after.stop_id)
        for arrive_time, arrive_stop_id in arrivals_zip:
            # time to reach is diff from start time to arrival (plus any baseline cost)
            arrive_time_adjusted = arrive_time - departure_secs + baseline_cost
            stops_state.try_add_update(
                arrive_stop_id,
                arrive_time_adjusted,
                trip_id,
                preceding_path)

# assume all xfers are 3 minutes
TRANSFER_COST = (5 * 60)

def add_footpath_transfers(
    stops_state: StopAccessState,
    stops_gdf: gpd.GeoDataFrame,
    already_processed_stops: List[str],
    transfer_cost=TRANSFER_COST,
) -> List[str]:
    # initialize a return object
    updated_stop_ids = []

    # add in transfers to nearby stops
    stop_ids = stops_state.all_stops()
    for stop_id in stop_ids:
        # no need to re-intersect already done stops
        if stop_id in already_processed_stops:
            continue

        stop_pt = stops_gdf.loc[stop_id].geometry

        # TODO: parameterize? transfer within .2 miles
        meters_in_miles = 1610
        qual_area = stop_pt.buffer(meters_in_miles/5)
        
        # get all stops within a short walk of target stop
        mask = stops_gdf.intersects(qual_area)

        # time to reach new nearby stops is the transfer cost plus arrival at last stop
        ref_stop_state = stops_state.get_stop(stop_id)
        arrive_time_adjusted = ref_stop_state["time_to_reach"] + TRANSFER_COST

        last_trip_id = None
        if len(ref_stop_state["preceding"]):
            last_trip_id = ref_stop_state["preceding"][-1]

        # only update if currently inaccessible or faster than currrent option
        for arrive_stop_id, row in stops_gdf[mask].iterrows():
            did_update = stops_state.try_add_update(
                arrive_stop_id,
                arrive_time_adjusted,
                last_trip_id)

            if did_update:
                updated_stop_ids.append(arrive_stop_id)
    
    return updated_stop_ids

import time

# initialize lookup with start node taking 0 seconds to reach
stop_access_state = StopAccessState(from_stop_id)

TRANSFER_LIMIT = 1
for k in range(TRANSFER_LIMIT + 1):
    print("\nAnalyzing possibilities with {} transfers".format(k))
    
    # generate current list of stop ids under consideration
    last_updated_stops = stop_access_state.all_stops()
    print("\tinital qualifying stop ids count: {}".format(len(last_updated_stops)))
    
    # update time to stops calculated based on stops accessible
    tic = time.perf_counter()
    stop_times_for_kth_trip(stop_access_state, last_updated_stops)
    toc = time.perf_counter()
    print("\tstop times calculated in {:0.4f} seconds".format(toc - tic))
    
    # now add footpath transfers and update
    tic = time.perf_counter()
    updated_stop_ids = add_footpath_transfers(stop_access_state, gdf, last_updated_stops)
    toc = time.perf_counter()
    print("\tfootpath transfers calculated in {:0.4f} seconds".format(toc - tic))
    print("\t\t{} stop ids added".format(len(updated_stop_ids)))
    
assert stop_access_state.has_stop(to_stop_id), "Unable to find route to destination within transfer limit"
print("Time to destination: {} minutes".format(stop_access_state.get_stop(to_stop_id)["time_to_reach"] / 60))